// Misc
function log(msg) {
    var now = new Date().toISOString().split("T").join(" ");
    var message = now + ": [JS] - " + msg;
    console.log(message);

    // on iOS using webkit callback handler
    // on Android using a custom defined LoggerHandler
    var handler = window.webkit && window.webkit.messageHandlers && window.webkit.messageHandlers.loggerHandler || LoggerHandler;
    if (handler) {
        handler.postMessage(message);
    }
}

var markersMovementAnimationRequest = null;
var animatingMarkers = new Map();

// Animation
function Animation() {
    this.line = null;
    this.lineDistance = null;
    this.duration = null;
    this.startTimeStamp = null;
}

// ClickableMarker
function ClickableMarker() {
    mmrgl.Marker.apply(this, arguments);

    this.id = null;
    this.animation = null;
    this.lastMovementDate = null;
}

ClickableMarker.prototype = Object.create(mmrgl.Marker.prototype);
ClickableMarker.prototype.constructor = ClickableMarker;
ClickableMarker.prototype.onClick = function(handleClick) {
    this._handleClick = handleClick;
    return this
}
ClickableMarker.prototype._onMapClick = function(e) {
    var targetElement = e.originalEvent.target;
    var element = this._element;

    if (this._handleClick && (targetElement === element || element.contains((targetElement)))) {
        this._handleClick();
    }
}

ClickableMarker.prototype.move = function(to, bearing, animated, duration) {
    if (to === undefined || to == null) {
        return;
    }

    // TODO: This is potential dangerous call. May be it should replace to `animatingMarkers.set(this.id, null)`
    animatingMarkers.delete(this.id)

    var fromCoordinate = this.getLngLat();
    var previousMovementDate = this.lastMovementDate;
    this.lastMovementDate = new Date();

    if (fromCoordinate === undefined || fromCoordinate == null || !animated) {
        this.setLngLat(to);
        this.addTo(controller.map);
        return;
    }

    var fromPoint = fromCoordinate.toArray();
    var toPoint = to;

    var animation = new Animation();

    animation.line = turf.lineString([fromPoint, toPoint]);
    animation.lineDistance = turf.distance(fromPoint, toPoint);

    if (duration === undefined || duration == null) {
        animation.duration = Math.min(this.lastMovementDate - previousMovementDate, 2000);
    } else {
        animation.duration = duration;
    }

    this.animation = animation;
    animatingMarkers.set(this.id, this);

//    log("moveMarkerAnimated. Id: " + this.id + ". Diff, time: " + animation.duration + "ms, distance: " + (1000 * animation.lineDistance).toFixed(2) + "m");

    if (markersMovementAnimationRequest) {
        cancelAnimationFrame(markersMovementAnimationRequest);
    }

    markersMovementAnimationRequest = requestAnimationFrame(animateMarkers);
    if (this === controller.currentLocationMarker && !this.isUserInteracting && controller.followLocation) {
        var mapBearing = controller.followBearing && bearing != null ? bearing : controller.map.getBearing();

        controller.map.flyTo({
        center: controller.followLocation ? to : null,
        bearing: mapBearing,
        duration: animation.duration,
        padding: mapPadding,
        animate: true,
            easing(t) {
                return t;
            }
        });
    }
}

function animateMarkers(timestamp) {
    var continueAnimation = false;

    animatingMarkers.forEach((marker, id) => {
        if (marker == null || marker.animation == null) {
            return;
        }

        var animation = marker.animation;

        if (animation.startTimeStamp == null) {
            animation.startTimeStamp = timestamp;
        }
        var timeDelta = timestamp - animation.startTimeStamp;
        var progress = timeDelta / animation.duration;
        var progressDistance = animation.lineDistance * progress;
        var alongPoint = turf.along(animation.line, progressDistance);
        var alongCoordinates = alongPoint.geometry.coordinates;

        marker.setLngLat(alongCoordinates);
        marker.addTo(controller.map);

        if (progress < 1.0) {
            continueAnimation = true;
        } else {
            animatingMarkers.set(id, null);
        }
    })

    if (continueAnimation) {
        markersMovementAnimationRequest = requestAnimationFrame(animateMarkers);
    }
}

// Controller
function Controller(map, marker) {
    this.map = map;
    this.currentLocationMarker = marker;
    this.markers = new Map();
    this.map.touchPitch.disable();
    this.map.keyboard.disable();

    this.currentCoordinate = null;
    this.currentBearing = null;
    this.followBearing = false;
    this.followLocation = false;
    this.isUserInteracting = false;
    this.isDriveModeOn = false;
}

Controller.prototype.zoomOut = function(step, animated) {
    this.map.flyTo({zoom: this.map.getZoom() - step, animate: animated, padding: mapPadding});
}

Controller.prototype.zoomIn = function(step, animated) {
    this.map.flyTo({zoom: this.map.getZoom() + step, animate: animated, padding: mapPadding});
}

Controller.prototype.setBearing = function(bearing, animated) {
    // Flag isUserInteracting will reset after "rotateEnd" map event
    this.isUserInteracting = true;
    this.map.flyTo({bearing: bearing, animate: animated, padding: mapPadding});
}

Controller.prototype.setZoom = function(zoom, animated) {
    this.map.flyTo({zoom: zoom, animate: animated, padding: mapPadding});
}

Controller.prototype.setMinMaxZoom = function(minZoom, maxZoom) {
    this.map.setMaxZoom(maxZoom)
    this.map.setMinZoom(minZoom)
}

Controller.prototype.setCenter = function(coords, animated) {
    // Flag isUserInteracting will reset after "moveEnd" map event
    this.isUserInteracting = true;
    this.map.flyTo({center: coords, animate: animated, padding: mapPadding});
}

Controller.prototype.setCenterAndDuration = function(coords, duration) {
    // Flag isUserInteracting will reset after "moveEnd" map event
    this.isUserInteracting = true;
    this.map.flyTo({center: coords, animate: true, duration: duration, curve: 1, padding: mapPadding});
}

Controller.prototype.setCenterWithZoom = function(coords, zoom, animated) {
    // Flag isUserInteracting will reset after "moveEnd" map event
    this.isUserInteracting = true;
    this.map.flyTo({center: coords, zoom: zoom, animate: animated, padding: mapPadding});
}

Controller.prototype.setCenterAndBearing = function(coords, bearing, animated) {
    // Flag isUserInteracting will reset after "moveEnd" map event
    this.isUserInteracting = true;
    this.map.flyTo({center: coords, bearing: bearing, animate: animated, padding: mapPadding});
}

Controller.prototype.markerProjection = function() {
    return this.map.project(this.currentLocationMarker.getLngLat());
}

Controller.prototype.fitBounds = function(bounds, animated) {
    this.map.fitBounds(bounds, {animate: animated, padding: mapPadding});
}

Controller.prototype.mapScale = function() {
    // A horizontal scale is imagined to be present at center of the map
    // container with maximum length (Default) as 100px.
    // Using spherical law of cosines approximation, the real distance is
    // found between the two coordinates.
    const maxWidth = 100;
    const y = this.map._container.clientHeight / 2;
    const left = this.map.unproject([0, y]);
    const right = this.map.unproject([maxWidth, y]);
    const maxMeters = left.distanceTo(right);
    return maxMeters / maxWidth;
}

Controller.prototype.setDriveMode = function(on) {
    this.isDriveModeOn = on;
    this.updateCurrentLocationMarker();
    this.setMapPadding();
}

Controller.prototype.setFollowBearing = function(on) {
    this.followBearing = on;
    this.setMapPadding();
}

Controller.prototype.setFollowLocation = function(on) {
    this.followLocation = on;
    this.setMapPadding();
}

Controller.prototype.setMapPadding = function() {
    if (!this.followLocation || !this.followBearing || !this.isDriveModeOn) {
        mapPadding = {top: 0, left: 0, right: 0, bottom: 0};
    } else {
        const height = this.map.getCanvas().clientHeight;
        mapPadding = {top: height/2};
    }

    this.map.flyTo({padding: mapPadding, animate: true});
}

Controller.prototype.sendEvent = function(event) {
    // on iOS using webkit callback handler
    // on Android using a custom defined MapsMessageHandler
    var handler = window.webkit && window.webkit.messageHandlers && window.webkit.messageHandlers.mapsMessageHandler || MapsMessageHandler;
    if (handler) {
        event.zoom = this.map.getZoom();
        event.bearing = this.map.getBearing();
        event.center = this.map.getCenter();
        event.bounds = this.map.getBounds();
        event.scale = this.mapScale();

        if (event.coords === undefined) {
            event.coords = event.center;
        }

        let project = this.map.project(event.coords);
        event.uiCoords = [project.x, project.y];

        handler.postMessage(JSON.stringify(event));
    }
}

Controller.prototype.setCurrentLocation = function(coords, bearing, accuracy) {
    if (coords) {
        this.currentLocationMarker.move(coords, bearing, true);
        currentLocation = coords;
    } else {
        currentLocation = [0, 0];
        this.currentLocationMarker.move([0, 0], 0, false);
    }

    currentLocationBearing = bearing;
    currentLocationAccuracy = accuracy;
    this.updateCurrentLocationBearing();
    this.updateCurrentLocationMarker();
}

Controller.prototype.updateCurrentLocationBearing = function() {
    if (!currentLocationBearing) return;
    var direction = currentLocationBearing - this.map.getBearing();
    var inner = document.querySelector('#marker_current_inner');
    inner.style.transform = 'rotate(' + direction + 'deg)';
}

Controller.prototype.updateCurrentLocationMarker = function() {
    var radius = 0;
    var bg = document.querySelector('#marker_current_bg');
    var point = document.querySelector('#marker_current_point');
    var arrow = document.querySelector('#marker_current_arrow');
    var direction = document.querySelector('#marker_current_direction');
    var drive = document.querySelector('#marker_direction_drive');

    if (currentLocation[1] && currentLocationAccuracy) {
        radius = pixelValue(currentLocation[1], currentLocationAccuracy, this.map.getZoom());
    } else {
        currentLocationAccuracy = 0;
    }

    bg.style.width = radius * 2 + 'px';
    bg.style.height = radius * 2 + 'px';

    if (this.isDriveModeOn) {
        bg.style.opacity = 0;
        point.style.opacity = 0;
        arrow.style.opacity = 0;
        direction.style.opacity = 0;
        drive.style.opacity = 1;
        return;
    }

    drive.style.opacity = 0;

    if (currentLocationAccuracy >= 50) {
        bg.style.opacity = 1;
        point.style.opacity = 0;
        arrow.style.opacity = 1;
        direction.style.opacity = 0;

    } else if (currentLocationAccuracy >= 15) {
        bg.style.opacity = 1;
        point.style.opacity = 1;
        arrow.style.opacity = 0;
        direction.style.opacity = 1;

    } else {
        bg.style.opacity = 0;
        point.style.opacity = 1;
        arrow.style.opacity = 0;
        direction.style.opacity = 1;
    }

    if (radius * 2 <= 24) { // center dot size
        bg.style.opacity = 0;
        point.style.opacity = 1;
        arrow.style.opacity = 0;
        direction.style.opacity = 1;
    }

    if (currentLocationBearing == null || currentLocationBearing == 'undefined') {
        arrow.style.opacity = 0;
        direction.style.opacity = 0;
    }

    if (currentLocation[0] == 0 && currentLocation[1] == 0) {
        point.style.opacity = 0;
    }
}

Controller.prototype.getCoordinates = function(screenCoords) {
    let project = this.map.unproject(screenCoords);
    controller.sendEvent({type: "getCoordinates", coords: project, uiCoords: screenCoords, userInitiated: true});
}

Controller.prototype.addMarker = function(id, coords, imageData, alignment = "center") {
    this.removeMarker(id);
    var marker = createMarker(id, coords, imageData, alignment);
    if (marker) {
        marker.addTo(this.map);
        this.markers.set(id, marker);
    }
}

Controller.prototype.removeMarker = function(id) {
    var marker = this.markers.get(id);
    if (marker) {
        marker.remove();
        this.markers.delete(id);
    }
}

Controller.prototype.removeAllMarkers = function() {
    this.markers.forEach(function(marker, index, map) {
        marker.remove();
    });
    this.markers.clear();
}

Controller.prototype.moveMarker = function(id, coords, animated, duration) {
    var marker = this.markers.get(id);
    if (marker) {
        marker.move(coords, null, animated, duration);
    }
}

Controller.prototype.displayPopup = function(id, content) {
    var marker = this.markers.get(id);
    if (marker) {
        var popupOffsets = {
            "top": [0, 0],
            "top-left": [0,0],
            "top-right": [0,0],
            "bottom": [0, -markerSize],
            "bottom-left": [0, -markerSize],
            "bottom-right": [0, -markerSize],
            "left": [markerSize / 2, -markerSize / 2],
            "right": [-markerSize / 2, -markerSize / 2]
        };

        marker.setPopup(new mmrgl.Popup({ offset: popupOffsets, focusAfterOpen: false }).setHTML(content));
        marker.getPopup().addTo(map);
    }
}

Controller.prototype.hidePopup = function(id) {
    var marker = this.markers.get(id);
    if (marker) {
        marker.getPopup().remove();
    }
}

Controller.prototype.changeStyle = function(style) {
    var style = "mmr://v2/styles/" + style + "_style.json";
    this.map.setStyle(style);
    log("Change style to: " + style);
}

Controller.prototype.setupCurrentLocationMarker = function() {
    if (this.currentLocationMarker) {
        this.currentLocationMarker.remove();
    }
    if (markerStyle) {
        markerStyle.remove();
    }

    var style = document.createElement("style");
    style.type = "text/css";
    style.innerHTML = `
    .marker_current-location {
                width: 48px;
                height: 48px;
                background-size: contain;
                background-repeat: no-repeat;
                background-position: center;
            }

            .marker__bg {
                background: rgba(0, 95, 249, .2);
                border: 1px solid #FFFFFF;
                width: 144px;
                height: 144px;
                border-radius: 50%;
                z-index: 1;
            }

            .marker__direction,
            .marker__bg,
            .marker__point {
                position: absolute;
                left: 50%;
                top: 50%;
                transform: translate(-50%, -50%);
                transition: all .2s;
            }

            .marker__point {
                background: rgb(0, 96, 250);
                width: 20px;
                height: 20px;
                border-radius: 50%;
                padding: 2px;
                z-index: 4
            }

            .marker__point:after {
                content: "";
                width: 100%;
                height: 100%;
                border-radius: 50%;
                border: 3px solid white;
                display: block;
                box-sizing: border-box;
            }

            .marker__direction {
                z-index: 2;
            }

            .marker__arrow {
                position: absolute;
                left: 50%;
                transform: translate(-50%, -90%);
                z-index: 3;
                transition: all .2s;
            }

            .marker__direction-inner {
                margin-bottom: 60px;
            }

            .marker__inner {
                transform: rotate(0deg);
            }

            .marker__drive {
                width: 92px;
                height: 92px;
                position: absolute;
                transform: translate(-50%, -50%);
                left: 50%;
                -webkit-filter: drop-shadow(0px 0px 2px rgba(0, 0, 0, 0.08)) drop-shadow(0px 0px 24px rgba(0, 0, 0, 0.16));
                filter: drop-shadow(0px 0px 2px rgba(0, 0, 0, 0.08)) drop-shadow(0px 0px 24px rgba(0, 0, 0, 0.16));
            }`;
    document.getElementsByTagName("head")[0].appendChild(style);
    markerStyle = style;

    if (this.currentLocationMarker) {
        this.currentLocationMarker.addTo(this.map);
    }
}

Controller.prototype.enableDragPan = function(enable) {
    if (enable) {
        this.map.dragPan.enable();
    } else {
        this.map.dragPan.disable();
    }
}

Controller.prototype.enableZoomRotate = function(enable) {
    if (enable) {
        this.map.touchZoomRotate.enable();
        this.map.doubleClickZoom.enable();
    } else {
        this.map.touchZoomRotate.disable();
        this.map.doubleClickZoom.disable();
    }
}

Controller.prototype.addSource = function(data, sourceId) {
    log("addSource ID: " + sourceId);

    var decodedData = window.atob(data);

    var dataJson;
    try {
        dataJson = JSON.parse(decodedData);
    } catch(e) {
        log("addSource failed to parse data: " + e)
    }
    if (!dataJson) { return; }
    this.map.addSource(sourceId + "-source", {
        'type': 'geojson',
        'data' : dataJson
    });
}

Controller.prototype.addPolylineSource = function(data, sourceId) {
    log("addPolylineSource ID: " + sourceId)

    var decodedData = window.atob(data);
    var coords = decodePolylineToCoords(decodedData);
    log("Coords length " + coords.length)

    const geoJSON = {
        type: 'FeatureCollection',
        features: [
            {
                type: 'Feature',
                geometry: {
                    type: 'LineString',
                    coordinates: coords
                }
            }
        ]
    };

    this.map.addSource(sourceId + "-source", {
        'type': 'geojson',
        'data' : geoJSON
    });
}

Controller.prototype.addCircleSource = function(coords, radius, steps, sourceId) {
    var options = {steps: steps, units: 'kilometers', properties: {}};
    var circle = turf.circle(coords, radius / 1000, options);

    this.map.addSource(sourceId + "-source", {
        'type': 'geojson',
        'data' : circle
    });
}

let clusterIDs = [];
const clusters = {};

Controller.prototype.addCluster = function(markersData, sourceId, radius, background, color) {
    log("addClusters ID: " + sourceId);

    var decodedData = window.atob(markersData);

    var markersJson;
    try {
        markersJson = JSON.parse(decodedData);
    } catch(e) {
        log("addClusters failed to parse data: " + e)
    }

    if (!markersJson) { return; }

    let features = [];
    for (const marker of markersJson) {
        features.push({
            type: 'Feature',
            properties: {
                id: marker.id,
                image: marker.image,
                alignment: marker.alignment
            },
            geometry: {
                type: 'Point',
                coordinates: marker.coords
            },
        });
    }

    const dataJson = {
        type: 'FeatureCollection',
        features: features
    };

    const id = sourceId + "-clusterSource";
    this.map.addSource(id, {
        'type': 'geojson',
        'data' : dataJson,
        'cluster': true,
        'clusterRadius': radius || 80
    });

    // add placeholder layer to get source features
    map.addLayer({
        'id': sourceId + '-clusterLayer',
        'type': 'line',
        'source': id
    });

    clusterIDs.push(id);
    clusters[id] = {
        markers: {},
        markersOnScreen: {},
        background: background || '#1f6ff9',
        color: color || 'white'
    };

    controller.updateClusterMarkers();
}

Controller.prototype.removeCluster = function(id) {
    try {
        let sourceId = id + "-clusterSource";
        for (const markerId in clusters[sourceId].markersOnScreen) {
            clusters[sourceId].markersOnScreen[markerId].remove();
            this.markers.delete(markerId);
        }

        clusterIDs.splice(clusterIDs.indexOf(sourceId), 1);
        delete clusters[sourceId];
        this.map.removeLayer(id + "-clusterLayer");
        this.map.removeSource(sourceId);
    } catch (e) {
        log("removeCluster failed, id: " + id + ", e: " + e);
    }
}

Controller.prototype.updateClusterMarkers = function() {
    for (const id of clusterIDs) {
        controller.updateClusterMarker(id);
    }
}

Controller.prototype.updateClusterMarker = function(sourceId) {
    const newMarkers = {};
    const features = map.querySourceFeatures(sourceId);

    for (const feature of features) {
        const coords = feature.geometry.coordinates;
        const props = feature.properties;
        const id = props.cluster ? props.cluster_id : props.id;
        const alignment = props.alignment || "center";

        let marker = clusters[sourceId].markers[id];

        if (!marker) {
            if (props.cluster) {
                const el = document.createElement('div');
                el.className = 'cluster';
                el.innerHTML = props.point_count;
                el.style.background = clusters[sourceId].background;
                el.style.color = clusters[sourceId].color;

                el.addEventListener('click', () => {
                      log("clickOnCluster, coords: " + coords);
                      controller.sendEvent({type: "clickOnCluster", coords: {lng: coords[0], lat: coords[1]}, userInitiated: true});
                   }
                );

                marker = clusters[sourceId].markers[id] = new mmrgl.Marker({
                    element: el
                }).setLngLat(coords);
            } else {
                marker = clusters[sourceId].markers[id] = createMarker(id, coords, props.image, alignment);
            }
        }

        newMarkers[id] = marker;

        if (!clusters[sourceId].markersOnScreen[id]) {
            marker.addTo(map);
            this.markers.set(id, marker);
        }
    }

    // Удаляем маркеры которые не видны больше
    for (const id in clusters[sourceId].markersOnScreen) {
        if (!newMarkers[id]) {
            clusters[sourceId].markersOnScreen[id].remove();
            this.markers.delete(id);
        }
    }
    clusters[sourceId].markersOnScreen = newMarkers;
}

Controller.prototype.addLayer = function(layerId, sourceId, type, paint) {
    log("addLayer ID: " + layerId + " for sourceId: " + sourceId);

    var paintJson;
    try {
        paintJson = JSON.parse(paint);
    } catch (e) {
        log("addLayer failed to parse paint: " + e)
    }
    if (!paintJson) { return; }

    this.map.addLayer({
        'id': layerId + "-layer",
        'type': type,
        'source': sourceId + "-source",
        'layout': type === 'line' ? {'line-join': 'round', 'line-cap': 'round'} : {},
        'paint': paintJson
    });
}

Controller.prototype.removeSource = function(sourceId) {
    try {
        this.map.removeSource(sourceId + "-source");
    } catch (e) {
        log("removeSource failed, sourceId: " + sourceId + ", e: " + e);
    }
}

Controller.prototype.removeLayer = function(layerId) {
    try {
        this.map.removeLayer(layerId + "-layer");
    } catch (e) {
        log("removeLayer failed, layerId: " + layerId + ", e: " + e);
    }
}

function createCurrentLocationMarker() {
    var el = document.createElement("div");
    el.id = "marker-container";
    el.className = "marker marker_current-location mmrgl-marker";
    el.innerHTML = `
        <div class="marker__inner" id="marker_current_inner">
                <div class="marker__bg" id="marker_current_bg">
                    <div class="marker__arrow" id="marker_current_arrow">
                        <svg width="14" height="11" viewBox="0 0 14 11" fill="none">
                            <path d="M6.56699 0.749999C6.75944 0.416667 7.24056 0.416666 7.43301 0.75L12.6292 9.75C12.8216 10.0833 12.5811 10.5 12.1962 10.5L1.80385 10.5C1.41895 10.5 1.17838 10.0833 1.37083 9.75L6.56699 0.749999Z" fill="#0060FA" stroke="white"></path>
                        </svg>
                    </div>
                </div>
                <div class="marker__point" id="marker_current_point"></div>
                <div class="marker__direction" id="marker_current_direction">
                    <svg class="marker__direction-inner" width="32" height="78" viewBox="0 0 32 78" fill="none" xmlns="http://www.w3.org/2000/svg">
                        <path d="M22 78L10 78L2.71777e-06 -1.82704e-07L32 -2.98023e-06L22 78Z" fill="url(#paint0_linear_707:10536)"/>
                        <defs>
                            <linearGradient id="paint0_linear_707:10536" x1="16.8078" y1="79" x2="17.5637" y2="0.564858" gradientUnits="userSpaceOnUse">
                                <stop stop-color="#005FF9" stop-opacity="0.4"/>
                                <stop offset="1" stop-color="white" stop-opacity="0"/>
                            </linearGradient>
                        </defs>
                    </svg>
                </div>
                <div class="marker__drive" id="marker_direction_drive">
                    <svg version="1.1" id="Слой_1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px" width="92" height="92"
                         viewBox="0 0 92 92" style="enable-background:new 0 0 92 92;" xml:space="preserve">
                    <style type="text/css">
                        .st0{fill:url(#SVGID_1_);}
                        .st1{fill:url(#SVGID_00000055686179169508369650000002211509808194058931_);}
                        .st2{fill:#FFFFFF;}
                    </style>
                    <linearGradient id="SVGID_1_" gradientUnits="userSpaceOnUse" x1="50.5308" y1="56.4405" x2="38.7434" y2="56.44" gradientTransform="matrix(1 0 0 -1 0 94)">
                        <stop  offset="0.2969" style="stop-color:#2688EB"/>
                        <stop  offset="0.4948" style="stop-color:#08409B"/>
                        <stop  offset="0.6615" style="stop-color:#08409B"/>
                    </linearGradient>
                    <path class="st0" d="M18,59.2l22.5-48.2c2.2-4.7,8.8-4.6,10.9,0L73,59.1c2.2,4.8-2.5,9.9-7.5,8.1l-17.7-6.3c-1.3-0.5-2.7-0.5-4,0
                        l-18.4,6.4C20.4,69.1,15.7,64,18,59.2z"/>
                    <linearGradient id="SVGID_00000004508148566712842020000000262490726586472107_" gradientUnits="userSpaceOnUse" x1="42.9192" y1="37.4705" x2="50.626" y2="37.5237" gradientTransform="matrix(1 0 0 -1 0 94)">
                        <stop  offset="0.2452" style="stop-color:#2688EB"/>
                        <stop  offset="0.5528" style="stop-color:#08409B"/>
                        <stop  offset="0.8212" style="stop-color:#08409B"/>
                    </linearGradient>
                    <path style="fill:url(#SVGID_00000004508148566712842020000000262490726586472107_);" d="M19.3,65.8l26-20.3c0.4-0.3,0.9-0.3,1.2,0
                        l25.3,20.3c0.9,0.7,0.1,2.1-0.9,1.7l-24.8-7.4c-0.2-0.1-0.4-0.1-0.6,0l-25.4,7.4C19.1,67.8,18.4,66.5,19.3,65.8z"/>
                    <g>
                        <g>
                            <path class="st2" d="M22.4,68.7c-1.7,0-3.2-0.7-4.4-2c-1.6-1.8-2-4.3-0.9-6.6l23.4-49.5c1-2.1,3.1-3.4,5.4-3.4c0,0,0,0,0,0
                                c2.4,0,4.4,1.4,5.4,3.5l22.5,49.4c1,2.2,0.6,4.7-1,6.5c-1.7,1.8-4.1,2.4-6.4,1.6l-18.7-6.6c-1.3-0.4-2.7-0.5-3.9,0l-19.4,6.7
                                C23.8,68.6,23.1,68.7,22.4,68.7z M19.8,61.5L19.8,61.5c-0.5,1.1-0.4,2.4,0.5,3.3c0.8,0.9,2,1.2,3.2,0.8l19.4-6.7
                                c1.9-0.7,4-0.6,5.9,0l18.7,6.6c1.2,0.4,2.4,0.1,3.2-0.8s1-2.1,0.5-3.3L48.6,12c-0.5-1.1-1.5-1.7-2.7-1.7c-1.2,0-2.2,0.6-2.7,1.7
                                L19.8,61.5z"/>
                        </g>
                    </g>
                    </svg>
                </div>
            </div>
            </div>`;
    let marker = new ClickableMarker(el);
    marker.setOffset([0, 24]);
    return marker;
}

function createMarker(id, coords, imageData, alignment = "center") {
    var el = document.createElement("div");
    el.id = "marker_" + id;
    el.className = "marker";

    if (imageData) {
        el.setAttribute("style", "background-image: url('data:image/png;base64," + imageData + "');");
    }

    var marker = new ClickableMarker(el);
    marker.id = id;
    marker.setLngLat(coords);
    if (alignment == "center") {
        marker.setOffset([0, 0]);
    } else if (alignment == "bottom") {
        marker.setOffset([0, -markerSize/2]);
    } else if (alignment == "top") {
        marker.setOffset([0, markerSize/2]);
    } else if (alignment == "right") {
        marker.setOffset([markerSize/2, 0]);
    } else if (alignment == "left") {
        marker.setOffset([-markerSize/2, 0]);
    }

    marker.onClick(function() {
        controller.sendEvent({type: "clickOnMarker", markerId: id, userInitiated: true});
    })

    return marker;
}

function createStyles() {
    var style = document.createElement("style");
    style.type = "text/css";
    style.innerHTML =
    ".marker {"
    + "width: " + markerSize + "px;"
    + "height: "+ markerSize + "px;"
    + "background-size: contain;"
    + "background-repeat: no-repeat;"
    + "background-position: center;"
    + "}";
    document.getElementsByTagName("head")[0].appendChild(style);
    createClusterStyle();
}

function createClusterStyle() {
    var style = document.createElement("style");
    style.type = "text/css";
    style.innerHTML = `
        .cluster {
            width: 50px;
            height: 50px;
            background: #1f6ff9;
            border-radius: 50%;
            color: white;
            display: flex;
            align-items: center;
            justify-content: center;
            font-size: 16px;
        }`;
    document.getElementsByTagName("head")[0].appendChild(style);
}

function injectTurf() {
    if (typeof turf === "undefined") {
        var script = document.createElement("script");
        script.setAttribute("src", "https://unpkg.com/@turf/turf@6/turf.min.js");
        document.getElementsByTagName("head")[0].appendChild(script);
    }
}

Controller.prototype.longTap = function(screenCoords) {
    let project = this.map.unproject(screenCoords);
    var isUserInitiated = true
    controller.sendEvent({type: "longPress", coords: project, userInitiated: true});
}

var markerSize = 48;
var markerStyle;
var currentLocationMarkerId = String(Date.now() + Math.random());
var currentLocationMarker = createCurrentLocationMarker();
currentLocationMarker.id = currentLocationMarkerId;
var currentLocationBearing = 0;
var currentLocation = [0, 0];
var currentLocationAccuracy = 0;
var mapPadding = {top: 0, left: 0, right: 0, bottom: 0};

createStyles();
injectTurf();

currentLocationMarker.onClick(function() {
    // Send "clickOnMarker" event to controller.
    // Right now, controller doesn't support click on current location marker.
})

var controller = new Controller(map, currentLocationMarker);

// Map event handlers

map.on("click", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "touch", coords: e.lngLat, userInitiated: typeof e.originalEvent !== "undefined"});
});

map.on("zoomstart", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.isUserInteracting = isUserInitiated
    controller.sendEvent({type: "zoomStart", userInitiated: isUserInitiated});
    controller.updateCurrentLocationMarker();
});

map.on("zoom", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "zoom", userInitiated: isUserInitiated});
    controller.updateCurrentLocationMarker();
});

map.on("zoomend", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "zoomEnd", userInitiated: isUserInitiated});
    controller.isUserInteracting = false
    controller.updateCurrentLocationMarker();
});

map.on("rotatestart", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.isUserInteracting = isUserInitiated
    controller.sendEvent({type: "rotateStart", userInitiated: isUserInitiated});
    controller.updateCurrentLocationBearing();
});

map.on("rotate", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "rotate", userInitiated: isUserInitiated});
    controller.updateCurrentLocationBearing();
});

map.on("rotateend", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "rotateEnd", userInitiated: isUserInitiated});
    controller.isUserInteracting = false
    controller.updateCurrentLocationBearing();
});

map.on("movestart", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.isUserInteracting = isUserInitiated
    controller.sendEvent({type: "moveStart", userInitiated: isUserInitiated});
});

map.on("move", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "move", userInitiated: isUserInitiated});
});

map.on("moveend", function(e) {
    var isUserInitiated = typeof e.originalEvent !== "undefined"
    controller.sendEvent({type: "moveEnd", userInitiated: isUserInitiated});
    controller.isUserInteracting = false
});

map.on('load', function(e) {
    controller.sendEvent({type: "load", userInitiated: false});
});

map.on('style.load', function(e) {
    controller.sendEvent({type: "styleLoad", userInitiated: false});
});

map.on('render', function(e) {
      controller.updateClusterMarkers();
});

// Helpers

var metersPerPixel = function(latitude, zoomLevel) {
      var earthCircumference = 40075017;
      var latitudeRadians = latitude * (Math.PI/180);
      return earthCircumference * Math.cos(latitudeRadians) / Math.pow(2, zoomLevel + 9);
};

var pixelValue = function(latitude, meters, zoomLevel) {
    return meters / metersPerPixel(latitude, zoomLevel);
};

// decode polyline

/**
 * decodePolylineToCoords
 * Декодирует строку в массив точек
 * @param {string} encodedString кодированная линия
 * @returns {[number, number][]} массив координат вида [lng, lat]
 * @type {(encodedString: string) => [number, number][]}
 */
function decodePolylineToCoords(encodedString) {
    const coordinates = [],
    factor = 1000000
    let index = 0,
    lat = 0,
    lng = 0,
    shift = 0,
    result = 0,
    byte = null,
    latitude_change,
    longitude_change;

    // Coordinates have variable length when encoded, so just keep
    // track of whether we've hit the end of the string. In each
    // loop iteration, a single coordinate is decoded.
    while (index < encodedString.length) {
        // Reset shift, result, and byte
        byte = null
        shift = 0
        result = 0

        do {
            byte = encodedString.charCodeAt(index++) - 63
            result |= (byte & 0x1f) << shift
            shift += 5
        } while (byte >= 0x20)

            latitude_change = result & 1 ? ~(result >> 1) : result >> 1

            shift = result = 0

            do {
                byte = encodedString.charCodeAt(index++) - 63
                result |= (byte & 0x1f) << shift
                shift += 5
            } while (byte >= 0x20)

                longitude_change = result & 1 ? ~(result >> 1) : result >> 1

                lat += latitude_change
                lng += longitude_change

                coordinates.push([lng / factor, lat / factor])
                }

    return coordinates
}

// load

function load() {
    return "ok";
}

load();